/*
 * SPI2CF (c) 2006, Till Harbaum.
 * All rights reserved.
 *
 * This file is distributed under the GPL.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS
 * OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
 * GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * This file is part SPI2CF port of the uIP TCP/IP stack.
 */

/*
  cf.c
*/

#include "uip-conf.h"

#include <stdio.h>
#include <avr/io.h>
#include <util/delay.h>

#include "spi.h"
#include "cf.h"

// keep track of gpos as they can't be read back
static uint16_t gpo_state;

void cf_set_gpo(uint16_t state) {
  gpo_state = state;
}

/* physically reset the card */
void cf_reset(void) {
  uint8_t i;

  gpo_state = 0;

  spi_start();
  spi_write_byte(CMD_BYTE(SET_ADDR));
  spi_write_word(0x8000);

  _delay_ms(10); 

  spi_start();
  spi_write_byte(CMD_BYTE(SET_ADDR));
  spi_write_word(0x0000);

  // delay 200ms (_delay_ms() is limited to 16ms @ 16Mhz)
  for(i=0;i<20;i++)
    _delay_ms(10); 

  spi_end();
}

void cf_get_id(void) {
  uint16_t id;

  spi_start();
  spi_write_byte(CMD_BYTE(GET_ID));
  id = spi_read_word() & 0x7fff;
  spi_end();

  printf_P(PSTR("SPI2CF chip id: %02x.%02x\n"), 
	 id>>8, id&0xff);
}

void cf_set_addr(uint16_t addr) {
  spi_start();
  spi_write_byte(CMD_BYTE(SET_ADDR));
  spi_write_word(gpo_state | addr);
  spi_end();
}

void cf_reg_read(uint16_t addr, char *dst, uint16_t bytes) {
  uint8_t result;

  cf_set_addr(addr);

  // and issue register read command
  spi_start();
  spi_write_byte(CMD_BYTE(REG_READ));
  spi_write_byte(0xff);
  while(bytes-- > 1)
    *dst++ = spi_read_byte();
  spi_end();

  // read last byte using special mechanism
  spi_start();
  *dst++ = spi_cycle_byte(CMD_BYTE(GET_ID));
  spi_end();
}

// read two bytes but make sure that the last byte does
// not cause the transfer of another one
uint16_t cf_read_last_word(void) {
  uint16_t result;
  uint8_t *p = (uint8_t*)&result;

  *p++ = spi_read_byte();
  spi_end();

  // read last byte using special mechanism
  spi_start();
  *p++ = spi_cycle_byte(CMD_BYTE(GET_ID));
  spi_end();

  return result;
}

uint8_t cf_reg_read_byte(uint16_t addr) {
  uint8_t byte;

  cf_set_addr(addr);

  spi_start();
  spi_write_byte(CMD_BYTE(REG_READ));
  spi_write_byte(0xff);
  byte = spi_cycle_byte(CMD_BYTE(GET_ID));
  spi_end();

  return byte;
}

uint16_t cf_reg_read_word(uint16_t addr) {
  cf_set_addr(CF_ADDR_BINC | addr);

  spi_start();
  spi_write_byte(CMD_BYTE(REG_READ));
  spi_write_byte(0xff);

  return cf_read_last_word();
}

void cf_reg_write_byte(uint16_t addr, uint8_t byte) {
  cf_set_addr(CF_ADDR_BINC | addr);

  spi_start();
  spi_write_byte(CMD_BYTE(REG_WRITE));
  spi_write_byte(byte);
  spi_write_byte(0xff);  // dummy byte will never be written
  spi_end();

  // disable CE etc by writing command
  spi_start();
  spi_write_byte(CMD_BYTE(GET_ID));
  spi_end();
}

uint16_t cf_io_read_word(uint16_t addr) {
  cf_set_addr(CF_ADDR_BINC | addr);

  spi_start();
  spi_write_byte(CMD_BYTE(IO_READ));
  spi_write_byte(0xff);
  return cf_read_last_word();
  // spi_end() is called within cf_read_last_word()
}

void cf_io_write_word(uint16_t addr, uint16_t val) {
  cf_set_addr(CF_ADDR_BINC | addr);

  spi_start();
  spi_write_byte(CMD_BYTE(IO_WRITE));
  spi_write_word(val);
  spi_write_byte(0xff);  // dummy byte will never be written
  spi_end();

  // disable CE etc by writing command
  spi_start();
  spi_write_byte(CMD_BYTE(GET_ID));
  spi_end();
}

void cf_io_read(uint16_t addr, uint16_t *dst, uint16_t words) {
  uint8_t result, *dbyte = (uint8_t*)dst;

  cf_set_addr(CF_ADDR_TOGGLE | addr);

  // and issue io read command
  spi_start();
  spi_write_byte(CMD_BYTE(IO_READ));
  spi_write_byte(0xff);
  while(words-- > 1) {
    *dbyte++ = spi_read_byte();
    *dbyte++ = spi_read_byte();
  }
  *dbyte++ = spi_read_byte();
  spi_end();

  // read last byte using special mechanism
  spi_start();
  *dbyte++ = spi_cycle_byte(CMD_BYTE(GET_ID));
  spi_end();
}

void cf_io_write(uint16_t addr, uint16_t *src, uint16_t words) {
  uint8_t result, *sbyte = (uint8_t*)src;

  cf_set_addr(CF_ADDR_TOGGLE | addr);

  // and issue io read command
  spi_start();
  spi_write_byte(CMD_BYTE(IO_WRITE));
  while(words--) {
    spi_write_byte(*sbyte++);
    spi_write_byte(*sbyte++);
  }

  spi_write_byte(0xff);  // dummy byte will never be written
  spi_end();

  // disable CE etc by writing command
  spi_start();
  spi_write_byte(CMD_BYTE(GET_ID));
  spi_end();
}

